; Useful utility routines for the 6809
; Alan Bleasby - December 2021
; Permission is given to freely use and distribute. Please give
; attribution.
; There is no warranty of any kind with this software

; ROM routines
INCH	equ	$8006		; Get KBD value in A (A=0 and 'Z' set if no
				; key is pressed)
OUTCH	equ	$800c		; Output character in A



; Non-ASCII key values
BREAK		equ	$03
SH_BREAK	equ	$03
UPARROW		equ	$5e
SH_UPARROW	equ	$5f
DOWNARROW	equ	$0a
SH_DOWNARROW	equ	$5b
SH_AT		equ	$13
LEFTARROW	equ	$08
SH_LEFTARROW	equ	$15
RIGHTARROW	equ	$09
SH_RIGHTARROW	equ	$5d
ENTER		equ	$0d
SH_ENTER	equ	$0d
CLEAR		equ	$0c
SH_CLEAR	equ	$5c

TIMER		equ	$112	; Source for random number seed


	section	"Utils"

	data
TNUM	rmb	$02		; PrtDec: value after valid subtractions
NUMDIV	rmb	$02		; Prt_Dec: Current power of 10 for subtracts
LEADZ	rmb	$01		; 0=No leading zeros, 1=leading zeros

NSYMS	equ	$04		; Number of symbols
HEXDIFF	equ	'A'-'9'-1	; Hex displacement in ASCII

INDEX	rmb	$01		; Index into string
ACCA	rmb	$04		; Main accumulator
ACCB	rmb	$04		; 2ry accumulator
	
BSYMS	fcc	" %@$",$00	; Valid base symbols for bin/oct/hex
ROLTAB	fcb	3,1,3,4		; Number of loops for dec/bin/oct/hex
BASETAB	fcb	10,2,8,16	; Max digit +1 for each base
BASELIM	fcb	0		; Storage for value from BASETAB
DECFLAG	fcb	0		; $ff if using decimal, else zero
NROLS	fcb	0		; Storage for value of ROLTAB
DIGIT	fcb	0		; Current digit being processed

TESTNUM	fcc	"6543",$00

BINBITS	equ	$10		; Number of bits in integer to convert
NBCDS	equ	$03		; Number of BCD bytes required for output

BINNUM	rmb	$02		; Space to hold 16 bit binary source
BCDNUM	rmb	NBCDS		; Space to hold BCD number

BINCNT	rmb	$01		; Counter for BINBITS
BCDTMP	rmb	$01		; BCD workspace

RANDOM	rmb	$02		; Seed & also last/latest random number


	code


; BinToBcd
; Converts an unsigned binary number to its equivalent packed BCD.
;
; INPUT
;   Binary number in location(s) BINNUM
; OUTPUT
;   Packed BCD right-justified in location(s) BCDNUM
;
; Requires the following to be set up before being called:
; BINBITS: The number of bits in the binary number 
;          The default is 16 to allow a range of $0000 - $FFFF
; NBCDS:   The number of BCD bytes required to hold an integer
;          of BINBITS bits
;          The default is 3 (allowing 65535) to be held
; For use with 32 bits you would set BINBITS=32 NBCDS=5 allowing a range of
; 0-4294967295.
; For use with 8 bits you would set BINBITS=8 NBCDS=2 allowing a range of
; 0-255.
; Alters: None
	
BinToBcd
	pshs	a,b,x,y
	ldb	#NBCDS-1
	leax	BCDNUM,pcr
	leay	BINNUM,pcr
1	clr	b,x
	decb
	bpl	1B

	ldb	#BINBITS
	stb	BINCNT,pcr

	; Check for nybbles >=5. Add 3 to any found
2	clrb
3	lda	b,x

	pshs	a
	lsra
	lsra
	lsra
	lsra
	cmpa	#$05
	bcs	4F
	adda	#$03
4	asla
	asla
	asla
	asla
	sta	BCDTMP,pcr
	puls	a
	anda	#$0f
	cmpa	#$05
	bcs	5F
	adda	#$03
5	adda	BCDTMP,pcr
	sta	b,x

	incb
	cmpb	#NBCDS
	bne	3B

	ldb	#(BINBITS/8)-1
6	rol	b,y
	decb
	bpl	6B

	ldb	#NBCDS-1
7	rol	b,x
	decb
	bpl	7B

	dec	BINCNT,pcr
	bne	2B
	puls	a,b,x,y
	rts




; BcdToBin
; Converts packed BCD to an unsigned int.
;
; INPUT
;   Packed BCD right-justified in location(s) BCDNUM
; OUTPUT
;   Binary number in location(s) BINNUM
;
; Requires the following to be set up before being called:
; BINBITS: The number of bits in the binary number 
;          The default is 16 to allow a range of $0000 - $FFFF
; NBCDS:   The number of BCD bytes required to hold an integer
;          of BINBITS bits
;          The default is 3 (allowing 65535) to be held
; For use with 32 bits you would set BINBITS=32 NBCDS=5 allowing a range of
; 0-4294967295.
; For use with 8 bits you would set BINBITS=8 NBCDS=2 allowing a range of
; 0-255.
; Alters: None
	
BcdToBin
	pshs	a,b,x,y
	ldb	#(BINBITS/8)-1
	leax	BCDNUM,pcr
	leay	BINNUM,pcr
1	clr	b,y
	decb
	bpl	1B
	
	ldb	#BINBITS
	stb	BINCNT,pcr


2	clrb
	lda	#NBCDS
3	ror	b,x
	incb
	deca
	bne	3B
	
	ldb	#$00		; clrb would clear carry
	lda	#(BINBITS/8)
4	ror	b,y
	incb
	deca
	bne	4B

	; Check for nybbles >=8. Subtract 3 from any found
	clrb
5	lda	b,x

	pshs	a
	lsra
	lsra
	lsra
	lsra
	cmpa	#$08
	bcs	6F
	suba	#$03
6	asla
	asla
	asla
	asla
	sta	BCDTMP,pcr
	puls	a
	anda	#$0f
	cmpa	#$08
	bcs	7F
	suba	#$03
7	adda	BCDTMP,pcr
	sta	b,x

	incb
	cmpb	#NBCDS
	bne	5B

	dec	BINCNT,pcr
	bne	2B
	puls	a,b,x,y
	rts




; PrtStr
; Prints a NULL terminated string.
;
; INPUT
;   Y points to the start of the string
; OUTPUT
;   Screen output at current cursor position
;
; Requires the following to be set up before being called:
; Load Y with the start address of the string
; Alters: Y

PrtStr
    	pshs	a
1	lda	,y+
	beq	2F
	jsr	OUTCH
	bra	1B
2	puls	a
	rts




; NumIn
; 16 bit numeric  input
;
; INPUT
;   Y points to the start of a numeric string
; OUTPUT
;   Integer stored in ACCA,pcr and Carry flag clear. On error the Carry
;   flag is set.
;
; Requires the following to be set up before being called.
; The input string must be NULL terminated and have no leading spaces.
; The default base is assumed to be decimal. Other bases  may be
; selected by prefixing the string with one of the following characters:
; $ = Hexadecimal
; % = Binary
; @ = Octal
; Alters: None
; Inspired by a 6502 routine by BigDumbDinosaur.
	
NumIn
	pshs	a,b,x,y
	clr	DECFLAG,pcr

	ldd	#$0000
	std	ACCA,pcr
	std	ACCA+2,pcr

	andcc	#%11111110
	lda	,y
	bne	1F
	puls	a,b,x,y
	rts


1	ldb	#NSYMS-1
	leax	BSYMS,pcr



2	cmpa	b,x
	beq	3F
	decb
	bpl	2B
	stb	DECFLAG
	incb


3	leax	BASETAB,pcr
	lda	b,x
	sta	BASELIM,pcr
	leax	ROLTAB,pcr	
	lda	b,x
	sta	NROLS,pcr

	cmpb	#$00
	beq	5F
	leay	1,y
	

4	clr	INDEX,pcr


5	andcc	#%11111110
	ldb	INDEX,pcr

	lda	b,y
	lbeq	12F
	inc	INDEX,pcr
	cmpa	#'a
	bcs	6F
	cmpa	#'z+1
	bcc	6F
	anda	#%01011111

6	suba	#'0
	lbcs	12F
	cmpa	#$0a
	bcs	7F
	suba	#HEXDIFF

7	cmpa	BASELIM,pcr
	lbcc	11F
	sta	DIGIT,pcr
	tst	DECFLAG,pcr
	bpl	8F


	lda	ACCA+3,pcr
	asla
	sta	ACCB+3,pcr
	lda	ACCA+2,pcr
	rola
	sta	ACCB+2,pcr
	lda	ACCA+1,pcr
	rola
	sta	ACCB+1,pcr
	lda	ACCA,pcr
	rola
	sta	ACCB,pcr

	lbcs	12F

8	ldb	NROLS,pcr
9	asl	ACCA+3,pcr
	rol	ACCA+2,pcr
	rol	ACCA+1,pcr
	rol	ACCA,pcr
	bcs	12F
	decb
	bne	9B

	tst	DECFLAG,pcr
	bpl	10F


	lda	ACCA+3,pcr
	adda	ACCB+3,pcr
	sta	ACCA+3,pcr
	lda	ACCA+2,pcr
	adca	ACCB+2,pcr
	sta	ACCA+2,pcr
	lda	ACCA+1,pcr
	adca	ACCB+1,pcr
	sta	ACCA+1,pcr
	lda	ACCA,pcr
	adca	ACCB,pcr
	sta	ACCA,pcr

	bcs	12F

10	lda	DIGIT,pcr
	adda	ACCA+3,pcr
	sta	ACCA+3,pcr
	lda	ACCA+2,pcr
	adca	#$00
	sta	ACCA+2,pcr
	lda	ACCA+1,pcr
	adca	#$00
	sta	ACCA+1,pcr
	lda	ACCA,pcr
	adca	#$00
	sta	ACCA,pcr

	bcs	12F

	ldb	INDEX,pcr
	lbpl	5B

11	orcc	#!00000001
12	puls	a,b,x,y
	rts

	


; ResetDp
; Sets the direct page to address $00
;
; INPUT
;   None
; OUTPUT
;   None
; Alters: None

ResetDp
	pshs	a
	clra
	tfr	a,dp
	setdp	$00
	puls	a
	rts




; GetKey
; Wait for a keypress and return character in A
;
; INPUT
;   None
; OUTPUT
;   None
; Alters: A

GetKey
	jsr	INCH
	beq	GetKey
	rts




; PrtHex
; Print byte as hex                  .
;
; INPUT
;   Byte in A
; OUTPUT
;   Screen hexadecimal number
; Alters: A

PrtHex
	pshs	a
	lsra
	lsra
	lsra
	lsra
	bsr	1F
	puls	a
1	anda	#$0f
	cmpa	#$0a
	bcs	2F
	adda	#$07
2	adda	#'0
	jmp	OUTCH





; PrtHex2
; Print two byte number as hex
;
; INPUT
;   Number in D
; OUTPUT
;   Screen hexadecimal number
; Alters: A,B

PrtHex2
	bsr	PrtHex
	tfr	b,a
	bsr	PrtHex
	rts




; PrtDec
; Print 16 bit unsigned int as decimal with or without leading zeros
;
; INPUT
;   16 bit unsigned integer in D
; OUTPUT
;   Screen decimal number
; Requirements:
;   If the byte LEADZ,pcr is non-zero then leading zeros will be output.
;   The field width in this case is 5 characters.
;   If the byte LEADZ,pcr is zero then no leading zeros will be supressed.
; Alters: None

PrtDec
	pshs	a,b,x,y
	cmpd	#$0000
	beq	4F

	ldx	#10000
	bsr	1F
	ldx	#1000
	bsr	1F
	ldx	#100
	bsr	1F
	ldx	#10
	bsr	1F
	ldx	#1
	bsr	1F
	puls	a,b,x,y
	rts

1	std	TNUM,pcr
	stx	NUMDIV,pcr
	ldx	#$0000

2	subd	NUMDIV,pcr
	bcs	3F
	leax	1,x		; Increment X
	bra	2B

3
	addd	NUMDIV,pcr
	std	TNUM,pcr

	tfr	x,d
	tfr	b,a



	ldb	LEADZ,pcr	; Print leading Zeros
	bne	4F		; Yes, so print char
	cmpa	#$00		; Yes, so is A zero 
	beq	5F		; Yes, so don't print
	dec	LEADZ,pcr	; An non-zero int so clear flag and print


4	adda	#'0
	jsr	OUTCH
5	ldd	TNUM,pcr
	rts




; Seed
; Set up the random number generator                  .
;
; INPUT
;   None
; OUTPUT
;   Copy system timer to RANDOM,pcr
; Alters: None

Seed
	pshs	a,b
	ldd	TIMER
	std	RANDOM,pcr
	puls	a,b
	rts




; RNG
; Pseudo random number generator. Period 65536
;
; INPUT
;   Last 16 bit random number from RANDOM,pcr
; OUTPUT
;   New 16 bit random number in RANDOM,pcr
; Alters: None
; Inspired by Veikko Sariola from an idea by George Marsaglia

RNG
	pshs	a
	lda	RANDOM,pcr
	lsra
	lda	RANDOM+1,pcr
	rora
	eora	RANDOM,pcr
	sta	RANDOM,pcr
	rora
	eora	RANDOM+1,pcr
	sta	RANDOM+1,pcr
	eora	RANDOM,pcr
	sta	RANDOM,pcr
	puls	a
	rts
